@@ -6,6 +6,7 @@
 #include <signal.h>
 #include "protobuf/core.pb-c.h"
 
+#include "asm-generic/page.h"
 #include "asm/bitops.h"
 #include "asm/int.h"
 
@@ -92,14 +93,6 @@ struct user_vfp_exc {
 #define ASSIGN_TYPED(a, b) do { a = (typeof(a))b; } while (0)
 #define ASSIGN_MEMBER(a,b,m) do { ASSIGN_TYPED((a)->m, (b)->m); } while (0)
 
-#ifndef PAGE_SIZE
-# define PAGE_SIZE	4096
-#endif
-
-#ifndef PAGE_MASK
-# define PAGE_MASK	(~(PAGE_SIZE - 1))
-#endif
-
 #define REG_RES(regs) ((regs).ARM_r0)
 #define REG_IP(regs)  ((regs).ARM_pc)
 #define REG_SYSCALL_NR(regs) ((regs).ARM_r7)
@@ -116,9 +109,9 @@ typedef UserArmRegsEntry UserRegsEntry;
 
 #define TI_SP(core) ((core)->ti_arm->gpregs->sp)
 
-typedef uint32_t auxv_t;
+typedef u32 auxv_t;
 
-static inline void *decode_pointer(uint64_t v) { return (void*)(uint32_t)v; }
-static inline uint64_t encode_pointer(void *p) { return (uint32_t)p; }
+static inline void *decode_pointer(u64 v) { return (void*)(u32)v; }
+static inline u64 encode_pointer(void *p) { return (u32)p; }
 
 #endif /* __CR_ASM_TYPES_H__ */
