@@ -70,9 +70,13 @@
 #include "vdso.h"
 #include "vma.h"
 #include "cr-service.h"
+#include "plugin.h"
+#include "irmap.h"
 
 #include "asm/dump.h"
 
+#define NR_ATTEMPTS 5
+
 static char loc_buf[PAGE_SIZE];
 
 bool privately_dump_vma(struct vma_area *vma)
@@ -80,15 +84,15 @@ bool privately_dump_vma(struct vma_area *vma)
 	/*
 	 * The special areas are not dumped.
 	 */
-	if (!(vma->vma.status & VMA_AREA_REGULAR))
+	if (!(vma->e->status & VMA_AREA_REGULAR))
 		return false;
 
 	/* No dumps for file-shared mappings */
-	if (vma->vma.status & VMA_FILE_SHARED)
+	if (vma->e->status & VMA_FILE_SHARED)
 		return false;
 
 	/* No dumps for SYSV IPC mappings */
-	if (vma->vma.status & VMA_AREA_SYSVIPC)
+	if (vma->e->status & VMA_AREA_SYSVIPC)
 		return false;
 
 	if (vma_area_is(vma, VMA_ANON_SHARED))
@@ -100,19 +104,32 @@ bool privately_dump_vma(struct vma_area *vma)
 		return false;
 	}
 
-	if (vma->vma.end > TASK_SIZE)
+	if (vma->e->end > TASK_SIZE)
 		return false;
 
 	return true;
 }
 
+static void close_vma_file(struct vma_area *vma)
+{
+	if (vma->vm_file_fd < 0)
+		return;
+	if (vma->e->status & VMA_AREA_SOCKET)
+		return;
+	if (vma->file_borrowed)
+		return;
+
+	close(vma->vm_file_fd);
+}
+
 void free_mappings(struct vm_area_list *vma_area_list)
 {
 	struct vma_area *vma_area, *p;
 
 	list_for_each_entry_safe(vma_area, p, &vma_area_list->h, list) {
-		if (vma_area->vm_file_fd > 0)
-			close(vma_area->vm_file_fd);
+		close_vma_file(vma_area);
+		if (!vma_area->file_borrowed)
+			free(vma_area->st);
 		free(vma_area);
 	}
 
@@ -132,7 +149,7 @@ int collect_mappings(pid_t pid, struct vm_area_list *vma_area_list)
 	if (ret < 0)
 		goto err;
 
-	pr_info("Collected, longest ares %lu bytes\n", vma_area_list->longest);
+	pr_info("Collected, longest area occupies %lu pages\n", vma_area_list->longest);
 	pr_info_vma_list(&vma_area_list->h);
 
 	pr_info("----------------------------------------\n");
@@ -227,7 +244,7 @@ static int collect_fds(pid_t pid, struct parasite_drain_fd *dfds)
 static int dump_task_exe_link(pid_t pid, MmEntry *mm)
 {
 	struct fd_parms params = FD_PARMS_INIT;
-	int fd, ret;
+	int fd, ret = 0;
 
 	fd = open_proc(pid, "exe");
 	if (fd < 0)
@@ -238,11 +255,10 @@ static int dump_task_exe_link(pid_t pid, MmEntry *mm)
 		return -1;
 	}
 
-	mm->exe_file_id = fd_id_generate_special();
+	if (fd_id_generate_special(&params.stat, &mm->exe_file_id))
+		ret = dump_one_reg_file(fd, mm->exe_file_id, &params);
 
-	ret = dump_one_reg_file(fd, mm->exe_file_id, &params);
 	close(fd);
-
 	return ret;
 }
 
@@ -264,11 +280,11 @@ static int dump_task_fs(pid_t pid, struct parasite_dump_misc *misc, struct cr_fd
 		return -1;
 	}
 
-	fe.cwd_id = fd_id_generate_special();
-
-	ret = dump_one_reg_file(fd, fe.cwd_id, &p);
-	if (ret < 0)
-		return ret;
+	if (fd_id_generate_special(&p.stat, &fe.cwd_id)) {
+		ret = dump_one_reg_file(fd, fe.cwd_id, &p);
+		if (ret < 0)
+			return ret;
+	}
 
 	close(fd);
 
@@ -282,11 +298,11 @@ static int dump_task_fs(pid_t pid, struct parasite_dump_misc *misc, struct cr_fd
 		return -1;
 	}
 
-	fe.root_id = fd_id_generate_special();
-
-	ret = dump_one_reg_file(fd, fe.root_id, &p);
-	if (ret < 0)
-		return ret;
+	if (fd_id_generate_special(&p.stat, &fe.root_id)) {
+		ret = dump_one_reg_file(fd, fe.root_id, &p);
+		if (ret < 0)
+			return ret;
+	}
 
 	close(fd);
 
@@ -326,23 +342,27 @@ static int dump_task_rlims(int pid, struct cr_fdset *fds)
 	return 0;
 }
 
-static int dump_filemap(pid_t pid, VmaEntry *vma, int file_fd,
+static int dump_filemap(pid_t pid, struct vma_area *vma_area,
 		const struct cr_fdset *fdset)
 {
 	struct fd_parms p = FD_PARMS_INIT;
+	VmaEntry *vma = vma_area->e;
+	int ret = 0;
+	u32 id;
 
-	if (fstat(file_fd, &p.stat) < 0) {
-		pr_perror("Can't stat file for vma");
-		return -1;
-	}
+	BUG_ON(!vma_area->st);
+	p.stat = *vma_area->st;
 
 	if ((vma->prot & PROT_WRITE) && vma_entry_is(vma, VMA_FILE_SHARED))
 		p.flags = O_RDWR;
 	else
 		p.flags = O_RDONLY;
-	vma->shmid = fd_id_generate_special();
 
-	return dump_one_reg_file(file_fd, vma->shmid, &p);
+	if (fd_id_generate_special(&p.stat, &id))
+		ret = dump_one_reg_file(vma_area->vm_file_fd, id, &p);
+
+	vma->shmid = id;
+	return ret;
 }
 
 static int check_sysvipc_map_dump(pid_t pid, VmaEntry *vma)
@@ -355,20 +375,54 @@ static int check_sysvipc_map_dump(pid_t pid, VmaEntry *vma)
 	return -1;
 }
 
-static int dump_task_mappings(pid_t pid, const struct vm_area_list *vma_area_list,
-			      const struct cr_fdset *cr_fdset)
+static int get_task_auxv(pid_t pid, MmEntry *mm)
 {
+	auxv_t mm_saved_auxv[AT_VECTOR_SIZE];
+	int fd, i, ret;
+
+	pr_info("Obtaining task auvx ...\n");
+
+	fd = open_proc(pid, "auxv");
+	if (fd < 0)
+		return -1;
+
+	ret = read(fd, mm_saved_auxv, sizeof(mm_saved_auxv));
+	if (ret < 0) {
+		ret = -1;
+		pr_perror("Error reading %d's auxv", pid);
+		goto err;
+	} else {
+		mm->n_mm_saved_auxv = ret / sizeof(auxv_t);
+		for (i = 0; i < mm->n_mm_saved_auxv; i++)
+			mm->mm_saved_auxv[i] = (u64)mm_saved_auxv[i];
+	}
+
+	ret = 0;
+err:
+	close_safe(&fd);
+	return ret;
+}
+
+static int dump_task_mm(pid_t pid, const struct proc_pid_stat *stat,
+		const struct parasite_dump_misc *misc,
+		const struct vm_area_list *vma_area_list,
+		const struct cr_fdset *fdset)
+{
+	MmEntry mme = MM_ENTRY__INIT;
 	struct vma_area *vma_area;
-	int ret = -1, fd;
+	int ret = -1, i = 0;
 
 	pr_info("\n");
-	pr_info("Dumping mappings (pid: %d)\n", pid);
+	pr_info("Dumping mm (pid: %d)\n", pid);
 	pr_info("----------------------------------------\n");
 
-	fd = fdset_fd(cr_fdset, CR_FD_VMAS);
+	mme.n_vmas = vma_area_list->nr;
+	mme.vmas = xmalloc(mme.n_vmas * sizeof(VmaEntry *));
+	if (!mme.vmas)
+		goto err;
 
 	list_for_each_entry(vma_area, &vma_area_list->h, list) {
-		VmaEntry *vma = &vma_area->vma;
+		VmaEntry *vma = vma_area->e;
 
 		pr_info_vma(vma_area);
 
@@ -380,20 +434,44 @@ static int dump_task_mappings(pid_t pid, const struct vm_area_list *vma_area_lis
 			ret = add_shmem_area(pid, vma);
 		else if (vma_entry_is(vma, VMA_FILE_PRIVATE) ||
 				vma_entry_is(vma, VMA_FILE_SHARED))
-			ret = dump_filemap(pid, vma, vma_area->vm_file_fd, cr_fdset);
+			ret = dump_filemap(pid, vma_area, fdset);
 		else if (vma_entry_is(vma, VMA_AREA_SOCKET))
 			ret = dump_socket_map(vma_area);
 		else
 			ret = 0;
-
-		if (!ret)
-			ret = pb_write_one(fd, vma, PB_VMA);
 		if (ret)
 			goto err;
+
+		mme.vmas[i++] = vma;
 	}
 
-	ret = 0;
-	pr_info("----------------------------------------\n");
+	mme.mm_start_code = stat->start_code;
+	mme.mm_end_code = stat->end_code;
+	mme.mm_start_data = stat->start_data;
+	mme.mm_end_data = stat->end_data;
+	mme.mm_start_stack = stat->start_stack;
+	mme.mm_start_brk = stat->start_brk;
+
+	mme.mm_arg_start = stat->arg_start;
+	mme.mm_arg_end = stat->arg_end;
+	mme.mm_env_start = stat->env_start;
+	mme.mm_env_end = stat->env_end;
+
+	mme.mm_brk = misc->brk;
+
+	mme.n_mm_saved_auxv = AT_VECTOR_SIZE;
+	mme.mm_saved_auxv = xmalloc(pb_repeated_size(&mme, mm_saved_auxv));
+	if (!mme.mm_saved_auxv)
+		goto err;
+
+	if (get_task_auxv(pid, &mme))
+		goto err;
+
+	if (dump_task_exe_link(pid, &mme))
+		goto err;
+
+	ret = pb_write_one(fdset_fd(fdset, CR_FD_MM), &mme, PB_MM);
+	xfree(mme.mm_saved_auxv);
 err:
 	return ret;
 }
@@ -434,74 +512,6 @@ static int dump_task_creds(struct parasite_ctl *ctl,
 	return pb_write_one(fdset_fd(fds, CR_FD_CREDS), &ce, PB_CREDS);
 }
 
-static int get_task_auxv(pid_t pid, MmEntry *mm, size_t *size)
-{
-	int fd, ret, i;
-
-	pr_info("Obtaining task auvx ... ");
-
-	fd = open_proc(pid, "auxv");
-	if (fd < 0)
-		return -1;
-
-	for (i = 0; i < AT_VECTOR_SIZE; i++) {
-		ret = read(fd, &mm->mm_saved_auxv[i],
-			   sizeof(auxv_t));
-		if (ret == 0)
-			break;
-		else if (ret != sizeof(auxv_t)) {
-			ret = -1;
-			pr_perror("Error reading %d's auxv[%d]",
-				  pid, i);
-			goto err;
-		}
-	}
-
-	*size = i;
-	ret = 0;
-err:
-	close_safe(&fd);
-	return ret;
-}
-
-static int dump_task_mm(pid_t pid, const struct proc_pid_stat *stat,
-		const struct parasite_dump_misc *misc, const struct cr_fdset *fdset)
-{
-	MmEntry mme = MM_ENTRY__INIT;
-	int ret = -1;
-
-	mme.mm_start_code = stat->start_code;
-	mme.mm_end_code = stat->end_code;
-	mme.mm_start_data = stat->start_data;
-	mme.mm_end_data = stat->end_data;
-	mme.mm_start_stack = stat->start_stack;
-	mme.mm_start_brk = stat->start_brk;
-
-	mme.mm_arg_start = stat->arg_start;
-	mme.mm_arg_end = stat->arg_end;
-	mme.mm_env_start = stat->env_start;
-	mme.mm_env_end = stat->env_end;
-
-	mme.mm_brk = misc->brk;
-
-	mme.n_mm_saved_auxv = AT_VECTOR_SIZE;
-	mme.mm_saved_auxv = xmalloc(pb_repeated_size(&mme, mm_saved_auxv));
-	if (!mme.mm_saved_auxv)
-		goto out;
-
-	if (get_task_auxv(pid, &mme, &mme.n_mm_saved_auxv))
-		goto out;
-	pr_info("OK\n");
-
-	if (dump_task_exe_link(pid, &mme))
-		goto out;
-
-	ret = pb_write_one(fdset_fd(fdset, CR_FD_MM), &mme, PB_MM);
-	xfree(mme.mm_saved_auxv);
-out:
-	return ret;
-}
-
 static int get_task_futex_robust_list(pid_t pid, ThreadCoreEntry *info)
 {
 	struct robust_list_head *head = NULL;
@@ -681,31 +691,6 @@ err:
 	return ret;
 }
 
-static int get_threads(struct pstree_item *item)
-{
-	return parse_threads(item->pid.real, &item->threads, &item->nr_threads);
-}
-
-static int check_threads(const struct pstree_item *item)
-{
-	struct pid *t = NULL;
-	int nr, ret;
-
-	ret = parse_threads(item->pid.real, &t, &nr);
-	if (ret)
-		return ret;
-
-	ret = ((nr == item->nr_threads) && !memcmp(t, item->threads, nr));
-	xfree(t);
-
-	if (!ret) {
-		pr_info("Threads set has changed while suspending\n");
-		return -1;
-	}
-
-	return 0;
-}
-
 static int parse_children(pid_t pid, pid_t **_c, int *_n)
 {
 	FILE *file;
@@ -756,37 +741,84 @@ err:
 	return -1;
 }
 
+static int collect_task(struct pstree_item *item);
 static int get_children(struct pstree_item *item)
 {
 	pid_t *ch;
-	int ret, i, nr_children;
+	int ret, i, nr_children, nr_inprogress;
 	struct pstree_item *c;
 
 	ret = parse_children(item->pid.real, &ch, &nr_children);
 	if (ret < 0)
 		return ret;
 
+	nr_inprogress = 0;
 	for (i = 0; i < nr_children; i++) {
+		pid_t pid = ch[i];
+
+		/* Is it already frozen? */
+		list_for_each_entry(c, &item->children, sibling)
+			if (c->pid.real == pid)
+				break;
+
+		if (&c->sibling != &item->children)
+			continue;
+
+		nr_inprogress++;
+
+		pr_info("Seized task %d, state %d\n", pid, ret);
+
 		c = alloc_pstree_item();
 		if (c == NULL) {
 			ret = -1;
 			goto free;
 		}
+
+		ret = seize_task(pid, item->pid.real, &item->pgid, &item->sid);
+		if (ret < 0) {
+			/*
+			 * Here is a race window between parse_children() and seize(),
+			 * so the task could die for these time.
+			 * Don't worry, will try again on the next attempt. The number
+			 * of attempts is restricted, so it will exit if something
+			 * really wrong.
+			 */
+			ret = 0;
+			xfree(c);
+			continue;
+		}
+
 		c->pid.real = ch[i];
 		c->parent = item;
+		c->state = ret;
 		list_add_tail(&c->sibling, &item->children);
+
+		/* Here is a recursive call (Depth-first search) */
+		ret = collect_task(c);
+		if (ret < 0)
+			goto free;
 	}
 free:
 	xfree(ch);
-	return ret;
+	return ret < 0 ? ret : nr_inprogress;
 }
 
 static void unseize_task_and_threads(const struct pstree_item *item, int st)
 {
 	int i;
 
-	for (i = 0; i < item->nr_threads; i++)
-		unseize_task(item->threads[i].real, st); /* item->pid will be here */
+	if (item->state == TASK_DEAD)
+		return;
+
+	/*
+	 * The st is the state we want to switch tasks into,
+	 * the item->state is the state task was in when we seized one.
+	 */
+
+	unseize_task(item->pid.real, item->state, st);
+
+	for (i = 1; i < item->nr_threads; i++)
+		ptrace(PTRACE_DETACH, item->threads[i].real, NULL, NULL);
 }
 
 static void pstree_switch_state(struct pstree_item *root_item, int st)
@@ -804,25 +836,58 @@ static pid_t item_ppid(const struct pstree_item *item)
 	return item ? item->pid.real : -1;
 }
 
-static int seize_threads(const struct pstree_item *item)
+static int seize_threads(struct pstree_item *item,
+				struct pid *threads, int nr_threads)
 {
-	int i = 0, ret;
+	int i = 0, ret, j, nr_inprogress;
 
-	if ((item->state == TASK_DEAD) && (item->nr_threads > 1)) {
+	if ((item->state == TASK_DEAD) && (nr_threads > 1)) {
 		pr_err("Zombies with threads are not supported\n");
 		goto err;
 	}
 
-	for (i = 0; i < item->nr_threads; i++) {
-		pid_t pid = item->threads[i].real;
+	/* The number of threads can't be less than allready frozen */
+	item->threads = xrealloc(item->threads, nr_threads * sizeof(struct pid));
+	if (item->threads == NULL)
+		return -1;
+
+	if (item->nr_threads == 0) {
+		item->threads[0].real = item->pid.real;
+		item->nr_threads = 1;
+	}
+
+	nr_inprogress = 0;
+	for (i = 0; i < nr_threads; i++) {
+		pid_t pid = threads[i].real;
 		if (item->pid.real == pid)
 			continue;
 
+		for (j = 0; j < item->nr_threads; j++)
+			if (pid == item->threads[j].real)
+				break;
+
+		if (j != item->nr_threads)
+			continue;
+		nr_inprogress++;
+
 		pr_info("\tSeizing %d's %d thread\n",
 				item->pid.real, pid);
+
 		ret = seize_task(pid, item_ppid(item), NULL, NULL);
-		if (ret < 0)
-			goto err;
+		if (ret < 0) {
+			/*
+			 * Here is a race window between parse_threads() and seize(),
+			 * so the task could die for these time.
+			 * Don't worry, will try again on the next attempt. The number
+			 * of attempts is restricted, so it will exit if something
+			 * really wrong.
+			 */
+			continue;
+		}
+
+		BUG_ON(item->nr_threads + 1 > nr_threads);
+		item->threads[item->nr_threads].real = pid;
+		item->nr_threads++;
 
 		if (ret == TASK_DEAD) {
 			pr_err("Zombie thread not supported\n");
@@ -835,50 +900,67 @@ static int seize_threads(const struct pstree_item *item)
 		}
 	}
 
-	return 0;
-
+	return nr_inprogress;
 err:
-	for (i--; i >= 0; i--) {
-		if (item->pid.real == item->threads[i].real)
-			continue;
-
-		unseize_task(item->threads[i].real, TASK_ALIVE);
-	}
-
 	return -1;
 }
 
 static int collect_threads(struct pstree_item *item)
 {
-	int ret;
+	int ret, attempts = NR_ATTEMPTS;
+	struct pid *t;
+	int nr, nr_inprogress;
 
-	ret = get_threads(item);
-	if (!ret)
-		ret = seize_threads(item);
-	if (!ret)
-		ret = check_threads(item);
+	nr_inprogress = 1;
+	while (nr_inprogress > 0 && attempts) {
+		attempts--;
 
-	return ret;
+		t = NULL;
+		nr = 0;
+
+		ret = parse_threads(item->pid.real, &t, &nr);
+		if (ret < 0)
+			break;
+
+		nr_inprogress = seize_threads(item, t, nr);
+		xfree(t);
+		if (nr_inprogress < 0)
+			break;
+
+	}
+
+	if (nr_inprogress && attempts)
+		return -1;
+
+	return 0;
 }
 
 static int collect_task(struct pstree_item *item)
 {
-	int ret;
-	pid_t pid = item->pid.real;
-
-	ret = seize_task(pid, item_ppid(item), &item->pgid, &item->sid);
-	if (ret < 0)
-		goto err;
-
-	pr_info("Seized task %d, state %d\n", pid, ret);
-	item->state = ret;
+	int ret, nr_inprogress, attempts = NR_ATTEMPTS;
 
 	ret = collect_threads(item);
 	if (ret < 0)
 		goto err_close;
 
-	ret = get_children(item);
-	if (ret < 0)
+	if (item->state == TASK_DEAD)
+		return 0;
+
+	/* Depth-first search (DFS) is used for traversing a process tree. */
+	nr_inprogress = 1;
+	while (nr_inprogress && attempts) {
+		attempts--;
+
+		/*
+		 * Freeze children and children of children, etc.
+		 * Then check again, that nobody is reparented.
+		 */
+		nr_inprogress = get_children(item);
+		if (nr_inprogress < 0)
+			goto err_close;
+	}
+
+	if (attempts == 0)
 		goto err_close;
 
 	if ((item->state == TASK_DEAD) && !list_empty(&item->children)) {
@@ -893,68 +975,9 @@ static int collect_task(struct pstree_item *item)
 
 err_close:
 	close_pid_proc();
-	unseize_task(pid, item->state);
-err:
 	return -1;
 }
 
-static int check_subtree(const struct pstree_item *item)
-{
-	pid_t *ch;
-	int nr, ret, i;
-	struct pstree_item *child;
-
-	ret = parse_children(item->pid.real, &ch, &nr);
-	if (ret < 0)
-		return ret;
-
-	i = 0;
-	list_for_each_entry(child, &item->children, sibling) {
-		if (child->pid.real != ch[i])
-			break;
-		i++;
-		if (i > nr)
-			break;
-	}
-	xfree(ch);
-
-	if (i != nr) {
-		pr_info("Children set has changed while suspending\n");
-		return -1;
-	}
-
-	return 0;
-}
-
-static int collect_subtree(struct pstree_item *item)
-{
-	struct pstree_item *child;
-	pid_t pid = item->pid.real;
-	int ret;
-
-	pr_info("Collecting tasks starting from %d\n", pid);
-	ret = collect_task(item);
-	if (ret)
-		return -1;
-
-	list_for_each_entry(child, &item->children, sibling) {
-		ret = collect_subtree(child);
-		if (ret < 0)
-			return -1;
-	}
-
-	/*
-	 * Tasks may clone() with the CLONE_PARENT flag while we collect
-	 * them, making more kids to their parent. So before proceeding
-	 * check that the parent we're working on has no more kids born.
-	 */
-
-	if (check_subtree(item))
-		return -1;
-
-	return 0;
-}
-
 static int collect_pstree_ids(void)
 {
 	struct pstree_item *item;
@@ -968,52 +991,32 @@ static int collect_pstree_ids(void)
 
 static int collect_pstree(pid_t pid)
 {
-	int ret, attempts = 5;
+	int ret;
 
 	timing_start(TIME_FREEZING);
 
-	while (1) {
-		root_item = alloc_pstree_item();
-		if (root_item == NULL)
-			return -1;
+	root_item = alloc_pstree_item();
+	if (root_item == NULL)
+		return -1;
 
-		root_item->pid.real = pid;
+	root_item->pid.real = pid;
+	ret = seize_task(pid, -1, &root_item->pgid, &root_item->sid);
+	if (ret < 0)
+		goto err;
+	pr_info("Seized task %d, state %d\n", pid, ret);
+	root_item->state = ret;
 
-		ret = collect_subtree(root_item);
-		if (ret == 0) {
-			/*
-			 * Some tasks could have been reparented to
-			 * namespaces' reaper. Check this.
-			 */
-			if (check_subtree(root_item))
-				goto try_again;
-
-			break;
-		}
-
-		/*
-		 * Old tasks can die and new ones can appear while we
-		 * try to seize the swarm. It's much simpler (and reliable)
-		 * just to restart the collection from the beginning
-		 * rather than trying to chase them.
-		 */
-try_again:
-		if (attempts == 0) {
-			pr_err("Can't freeze the tree\n");
-			return -1;
-		}
-
-		attempts--;
-		pr_info("Trying to suspend tasks again\n");
-
-		pstree_switch_state(root_item, TASK_ALIVE);
-		free_pstree(root_item);
-	}
+	ret = collect_task(root_item);
+	if (ret < 0)
+		goto err;
 
 	timing_stop(TIME_FREEZING);
 	timing_start(TIME_FROZEN);
 
 	return 0;
+err:
+	pstree_switch_state(root_item, TASK_ALIVE);
+	return -1;
 }
 
 static int collect_file_locks(void)
@@ -1338,12 +1341,24 @@ static int pre_dump_one_task(struct pstree_item *item, struct list_head *ctls)
 		goto err_free;
 	}
 
+	ret = parasite_fixup_vdso(parasite_ctl, pid, &vmas);
+	if (ret) {
+		pr_err("Can't fixup vdso VMAs (pid: %d)\n", pid);
+		goto err_cure;
+	}
+
 	ret = parasite_dump_misc_seized(parasite_ctl, &misc);
 	if (ret) {
 		pr_err("Can't dump misc (pid: %d)\n", pid);
 		goto err_cure;
 	}
 
+	ret = predump_task_files(pid);
+	if (ret) {
+		pr_err("Pre-dumping files failed (pid: %d)\n", pid);
+		goto err_cure;
+	}
+
 	parasite_ctl->pid.virt = item->pid.virt = misc.pid;
 
 	ret = parasite_dump_pages_seized(parasite_ctl, &vmas, &parasite_ctl->mem_pp);
@@ -1421,7 +1436,7 @@ static int dump_one_task(struct pstree_item *item)
 	}
 
 	ret = parse_posix_timers(pid, &proc_args);
-	if (ret < 0){
+	if (ret < 0) {
 		pr_err("Can't read posix timers file (pid: %d)\n", pid);
 		goto err;
 	}
@@ -1524,12 +1539,6 @@ static int dump_one_task(struct pstree_item *item)
 		goto err_cure;
 	}
 
-	ret = dump_task_mm(pid, &pps_buf, &misc, cr_fdset);
-	if (ret) {
-		pr_err("Dump mm (pid: %d) failed with %d\n", pid, ret);
-		goto err_cure;
-	}
-
 	ret = dump_task_core_all(item, &pps_buf, &misc, cr_fdset);
 	if (ret) {
 		pr_err("Dump core (pid: %d) failed with %d\n", pid, ret);
@@ -1554,7 +1563,7 @@ static int dump_one_task(struct pstree_item *item)
 		goto err;
 	}
 
-	ret = dump_task_mappings(pid, &vmas, cr_fdset);
+	ret = dump_task_mm(pid, &pps_buf, &misc, &vmas, cr_fdset);
 	if (ret) {
 		pr_err("Dump mappings (pid: %d) failed with %d\n", pid, ret);
 		goto err;
@@ -1600,18 +1609,46 @@ int cr_pre_dump_tasks(pid_t pid)
 	LIST_HEAD(ctls);
 	struct parasite_ctl *ctl, *n;
 
+	if (!opts.track_mem) {
+		pr_info("Enforcing memory tracking for pre-dump.\n");
+		opts.track_mem = true;
+	}
+
+	if (opts.final_state == TASK_DEAD) {
+		pr_info("Enforcing tasks run after pre-dump.\n");
+		opts.final_state = TASK_ALIVE;
+	}
+
 	if (init_stats(DUMP_STATS))
 		goto err;
 
 	if (kerndat_init())
 		goto err;
 
+	if (irmap_load_cache())
+		goto err;
+
+	if (cpu_init())
+		goto err;
+
+	if (vdso_init())
+		goto err;
+
 	if (connect_to_page_server())
 		goto err;
 
 	if (collect_pstree(pid))
 		goto err;
 
+	if (gen_predump_ns_mask())
+		goto err;
+
+	if (collect_mount_info(pid))
+		goto err;
+
+	if (mntns_collect_root(root_item->pid.real))
+		goto err;
+
 	for_each_pstree_item(item)
 		if (pre_dump_one_task(item, &ctls))
 			goto err;
@@ -1637,6 +1674,10 @@ err:
 		ret = page_xfer_dump_pages(&xfer, ctl->mem_pp, 0);
 
 		xfer.close(&xfer);
+
+		if (ret)
+			break;
+
 		timing_stop(TIME_MEMWRITE);
 
 		destroy_page_pipe(ctl->mem_pp);
@@ -1644,6 +1685,9 @@ err:
 		parasite_cure_local(ctl);
 	}
 
+	if (irmap_predump_run())
+		ret = -1;
+
 	if (disconnect_from_page_server())
 		ret = -1;
 
@@ -1670,9 +1714,15 @@ int cr_dump_tasks(pid_t pid)
 	if (init_stats(DUMP_STATS))
 		goto err;
 
+	if (cr_plugin_init())
+		goto err;
+
 	if (kerndat_init())
 		goto err;
 
+	if (irmap_load_cache())
+		goto err;
+
 	if (cpu_init())
 		goto err;
 
@@ -1753,6 +1803,8 @@ err:
 
 	close_cr_fdset(&glob_fdset);
 
+	cr_plugin_fini();
+
 	if (!ret) {
 		/*
 		 * It might be a migration case, where we're asked
@@ -1814,5 +1866,5 @@ err:
 		pr_info("Dumping finished successfully\n");
 	}
 
-	return post_dump_ret ? : ret;
+	return post_dump_ret ? : (ret != 0);
 }
