@@ -14,9 +14,10 @@
 
 #include "asm/types.h"
 #include "list.h"
+#include "fdset.h"
 #include "namespaces.h"
 #include "compiler.h"
-#include "crtools.h"
+#include "cr_options.h"
 #include "util.h"
 #include "sockets.h"
 #include "image.h"
@@ -24,6 +25,7 @@
 #include "ipc_ns.h"
 #include "pstree.h"
 #include "cr-show.h"
+#include "crtools.h"
 
 #include "protobuf.h"
 #include "protobuf/pstree.pb-c.h"
@@ -317,16 +319,18 @@ static struct show_image_info show_infos[] = {
 	SHOW_PLAINS(FILE_LOCK),
 	SHOW_PLAIN(RLIMIT),
 	SHOW_PLAIN(TUNFILE),
+	SHOW_PLAINS(EXT_FILE),
+	SHOW_PLAIN(IRMAP_CACHE),
 
-	{ TCP_STREAM_MAGIC,	PB_TCP_STREAM,		true,	show_tcp_stream, "1:%u 2:%u 3:%u 4:%u", },
-	{ STATS_MAGIC,		PB_STATS,		true,	NULL, "1.1:%u 1.2:%u 1.3:%u 1.4:%u 1.5:%Lu 1.6:%Lu 1.7:%Lu", },
+	{ TCP_STREAM_MAGIC,	PB_TCP_STREAM,		true,	show_tcp_stream, "1:%u 2:%u 3:%u 4:%u 12:%u", },
+	{ STATS_MAGIC,		PB_STATS,		true,	NULL, "1.1:%u 1.2:%u 1.3:%u 1.4:%u 1.5:%Lu 1.6:%Lu 1.7:%Lu 1.8:%u", },
 	{ FDINFO_MAGIC,		PB_FDINFO,		false,	NULL, "flags:%#o fd:%d", },
 	{ UNIXSK_MAGIC,		PB_UNIX_SK,		false,	NULL, "1:%#x 2:%#x 3:%d 4:%d 5:%d 6:%d 7:%d 8:%#x 11:S", },
 	{ INETSK_MAGIC,		PB_INET_SK,		false,	NULL, "1:%#x 2:%#x 3:%d 4:%d 5:%d 6:%d 7:%d 8:%d 9:%2x 11:A 12:A", },
 	{ PACKETSK_MAGIC,	PB_PACKET_SOCK,		false,	NULL, "5:%d", },
 	{ ITIMERS_MAGIC,	PB_ITIMER,		false,	NULL, "*:%Lu", },
 	{ POSIX_TIMERS_MAGIC,	PB_POSIX_TIMER,		false,	NULL, "*:%d 5:%Lu 7:%Lu 8:%lu 9:%Lu 10:%Lu", },
-	{ NETDEV_MAGIC,		PB_NETDEV, 		false,	NULL, "2:%d", },
+	{ NETDEV_MAGIC,		PB_NETDEV,		false,	NULL, "2:%d", },
 
 	{ PAGEMAP_MAGIC,	PB_PAGEMAP_HEAD,	true,	show_pagemaps,		NULL, },
 	{ PIPES_DATA_MAGIC,	PB_PIPE_DATA,		false,	pipe_data_handler,	NULL, },
@@ -475,7 +479,7 @@ static int cr_show_pid(int pid)
 		PstreeEntry *pe;
 
 		ret = pb_read_one_eof(fd, &pe, PB_PSTREE);
-		if (ret <= 0){
+		if (ret <= 0) {
 			close(fd);
 			return ret;
 		}
