@@ -16,6 +16,8 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 
+#include <dlfcn.h>
+
 #include "asm/types.h"
 
 #include "compiler.h"
@@ -31,6 +33,7 @@
 #include "tty.h"
 #include "file-lock.h"
 #include "cr-service.h"
+#include "plugin.h"
 
 struct cr_options opts;
 
@@ -68,7 +71,7 @@ static int parse_ns_string(const char *ptr)
 	return 0;
 
 bad_ns:
-	pr_err("Unknown namespace '%s'\n", ptr);
+	pr_msg("Error: unknown namespace: %s\n", ptr);
 	return -1;
 }
 
@@ -76,10 +79,50 @@ int main(int argc, char *argv[])
 {
 	pid_t pid = 0, tree_id = 0;
 	int ret = -1;
+	bool usage_error = true;
 	int opt, idx;
-	int log_level = 0;
+	int log_level = LOG_UNSET;
 	char *imgs_dir = ".";
 	char *work_dir = NULL;
+	static const char short_opts[] = "dsRf:F:t:p:hcD:o:n:v::xVr:jlW:L:";
+	static struct option long_opts[] = {
+		{ "tree", required_argument, 0, 't' },
+		{ "pid", required_argument, 0, 'p' },
+		{ "leave-stopped", no_argument, 0, 's' },
+		{ "leave-running", no_argument, 0, 'R' },
+		{ "restore-detached", no_argument, 0, 'd' },
+		{ "daemon", no_argument, 0, 'd' },
+		{ "contents", no_argument, 0, 'c' },
+		{ "file", required_argument, 0, 'f' },
+		{ "fields", required_argument, 0, 'F' },
+		{ "images-dir", required_argument, 0, 'D' },
+		{ "work-dir", required_argument, 0, 'W' },
+		{ "log-file", required_argument, 0, 'o' },
+		{ "namespaces", required_argument, 0, 'n' },
+		{ "root", required_argument, 0, 'r' },
+		{ USK_EXT_PARAM, no_argument, 0, 'x' },
+		{ "help", no_argument, 0, 'h' },
+		{ SK_EST_PARAM, no_argument, 0, 42 },
+		{ "close", required_argument, 0, 43 },
+		{ "log-pid", no_argument, 0, 44},
+		{ "version", no_argument, 0, 'V'},
+		{ "evasive-devices", no_argument, 0, 45},
+		{ "pidfile", required_argument, 0, 46},
+		{ "veth-pair", required_argument, 0, 47},
+		{ "action-script", required_argument, 0, 49},
+		{ LREMAP_PARAM, no_argument, 0, 41},
+		{ OPT_SHELL_JOB, no_argument, 0, 'j'},
+		{ OPT_FILE_LOCKS, no_argument, 0, 'l'},
+		{ "page-server", no_argument, 0, 50},
+		{ "address", required_argument, 0, 51},
+		{ "port", required_argument, 0, 52},
+		{ "prev-images-dir", required_argument, 0, 53},
+		{ "ms", no_argument, 0, 54},
+		{ "track-mem", no_argument, 0, 55},
+		{ "auto-dedup", no_argument, 0, 56},
+		{ "libdir", required_argument, 0, 'L'},
+		{ },
+	};
 
 	BUILD_BUG_ON(PAGE_SIZE != PAGE_IMAGE_SIZE);
 
@@ -92,46 +135,10 @@ int main(int argc, char *argv[])
 	init_opts();
 
 	if (init_service_fd())
-		return -1;
+		return 1;
 
 	while (1) {
-		static const char short_opts[] = "dsRf:t:p:hcD:o:n:v::xVr:jlW:";
-		static struct option long_opts[] = {
-			{ "tree", required_argument, 0, 't' },
-			{ "pid", required_argument, 0, 'p' },
-			{ "leave-stopped", no_argument, 0, 's' },
-			{ "leave-running", no_argument, 0, 'R' },
-			{ "restore-detached", no_argument, 0, 'd' },
-			{ "daemon", no_argument, 0, 'd' },
-			{ "contents", no_argument, 0, 'c' },
-			{ "file", required_argument, 0, 'f' },
-			{ "images-dir", required_argument, 0, 'D' },
-			{ "work-dir", required_argument, 0, 'W' },
-			{ "log-file", required_argument, 0, 'o' },
-			{ "namespaces", required_argument, 0, 'n' },
-			{ "root", required_argument, 0, 'r' },
-			{ USK_EXT_PARAM, no_argument, 0, 'x' },
-			{ "help", no_argument, 0, 'h' },
-			{ SK_EST_PARAM, no_argument, 0, 42 },
-			{ "close", required_argument, 0, 43 },
-			{ "log-pid", no_argument, 0, 44},
-			{ "version", no_argument, 0, 'V'},
-			{ "evasive-devices", no_argument, 0, 45},
-			{ "pidfile", required_argument, 0, 46},
-			{ "veth-pair", required_argument, 0, 47},
-			{ "action-script", required_argument, 0, 49},
-			{ LREMAP_PARAM, no_argument, 0, 41},
-			{ OPT_SHELL_JOB, no_argument, 0, 'j'},
-			{ OPT_FILE_LOCKS, no_argument, 0, 'l'},
-			{ "page-server", no_argument, 0, 50},
-			{ "address", required_argument, 0, 51},
-			{ "port", required_argument, 0, 52},
-			{ "prev-images-dir", required_argument, 0, 53},
-			{ "ms", no_argument, 0, 54},
-			{ "track-mem", no_argument, 0, 55},
-			{ },
-		};
-
+		idx = -1;
 		opt = getopt_long(argc, argv, short_opts, long_opts, &idx);
 		if (opt == -1)
 			break;
@@ -148,9 +155,13 @@ int main(int argc, char *argv[])
 			break;
 		case 'p':
 			pid = atoi(optarg);
+			if (pid <= 0)
+				goto bad_arg;
 			break;
 		case 't':
 			tree_id = atoi(optarg);
+			if (tree_id <= 0)
+				goto bad_arg;
 			break;
 		case 'c':
 			opts.show_pages_content	= true;
@@ -158,6 +169,9 @@ int main(int argc, char *argv[])
 		case 'f':
 			opts.show_dump_file = optarg;
 			break;
+		case 'F':
+			opts.show_fmt = optarg;
+			break;
 		case 'r':
 			opts.root = optarg;
 			break;
@@ -175,9 +189,11 @@ int main(int argc, char *argv[])
 			break;
 		case 'n':
 			if (parse_ns_string(optarg))
-				return -1;
+				goto bad_arg;
 			break;
 		case 'v':
+			if (log_level == LOG_UNSET)
+				log_level = 0;
 			if (optarg) {
 				if (optarg[0] == 'v')
 					/* handle -vvvvv */
@@ -214,21 +230,15 @@ int main(int argc, char *argv[])
 			break;
 		case 47:
 			{
-				struct veth_pair *n;
+				char *aux;
 
-				n = xmalloc(sizeof(*n));
-				if (n == NULL)
-					return -1;
-				n->outside = strchr(optarg, '=');
-				if (n->outside == NULL) {
-					xfree(n);
-					pr_err("Invalid argument for --veth-pair\n");
-					goto usage;
-				}
+				aux = strchr(optarg, '=');
+				if (aux == NULL)
+					goto bad_arg;
 
-				*n->outside++ = '\0';
-				n->inside = optarg;
-				list_add(&n->node, &opts.veth_pairs);
+				*aux = '\0';
+				if (veth_pair_add(optarg, aux + 1))
+					return 1;
 			}
 			break;
 		case 49:
@@ -237,7 +247,7 @@ int main(int argc, char *argv[])
 
 				script = xmalloc(sizeof(struct script));
 				if (script == NULL)
-					return -1;
+					return 1;
 
 				script->path = optarg;
 				list_add(&script->node, &opts.scripts);
@@ -251,10 +261,8 @@ int main(int argc, char *argv[])
 			break;
 		case 52:
 			opts.ps_port = htons(atoi(optarg));
-			if (!opts.ps_port) {
-				pr_err("Bad port\n");
-				return -1;
-			}
+			if (!opts.ps_port)
+				goto bad_arg;
 			break;
 		case 'j':
 			opts.shell_job = true;
@@ -268,15 +276,23 @@ int main(int argc, char *argv[])
 		case 55:
 			opts.track_mem = true;
 			break;
+		case 56:
+			opts.auto_dedup = true;
+			break;
 		case 54:
 			opts.check_ms_kernel = true;
 			break;
+		case 'L':
+			opts.libdir = optarg;
+			break;
 		case 'V':
 			pr_msg("Version: %s\n", CRIU_VERSION);
 			if (strcmp(CRIU_GITID, "0"))
 				pr_msg("GitID: %s\n", CRIU_GITID);
 			return 0;
 		case 'h':
+			usage_error = false;
+			goto usage;
 		default:
 			goto usage;
 		}
@@ -285,27 +301,27 @@ int main(int argc, char *argv[])
 	if (work_dir == NULL)
 		work_dir = imgs_dir;
 
-	if (optind >= argc)
+	if (optind >= argc) {
+		pr_msg("Error: command is required\n");
 		goto usage;
+	}
 
 	/* We must not open imgs dir, if service is called */
 	if (strcmp(argv[optind], "service")) {
 		ret = open_image_dir(imgs_dir);
-		if (ret < 0) {
-			pr_perror("Can't open imgs directory");
-			return -1;
-		}
+		if (ret < 0)
+			return 1;
 	}
 
 	if (chdir(work_dir)) {
 		pr_perror("Can't change directory to %s", work_dir);
-		return -1;
+		return 1;
 	}
 
 	log_set_loglevel(log_level);
 
 	if (log_init(opts.output))
-		return -1;
+		return 1;
 
 	if (opts.img_parent)
 		pr_info("Will do snapshot from %s\n", opts.img_parent);
@@ -320,46 +336,39 @@ int main(int argc, char *argv[])
 		if (!tree_id)
 			goto opt_pid_missing;
 
-		if (!opts.track_mem) {
-			pr_info("Enforcing memory tracking for pre-dump.\n");
-			opts.track_mem = true;
-		}
-
-		if (opts.final_state == TASK_DEAD) {
-			pr_info("Enforcing tasks run after pre-dump.\n");
-			opts.final_state = TASK_ALIVE;
-		}
-
-		return cr_pre_dump_tasks(tree_id);
+		return cr_pre_dump_tasks(tree_id) != 0;
 	}
 
 	if (!strcmp(argv[optind], "restore")) {
 		if (tree_id)
 			pr_warn("Using -t with criu restore is obsoleted\n");
-		return cr_restore_tasks();
+		return cr_restore_tasks() != 0;
 	}
 
 	if (!strcmp(argv[optind], "show"))
-		return cr_show(pid);
+		return cr_show(pid) != 0;
 
 	if (!strcmp(argv[optind], "check"))
-		return cr_check();
+		return cr_check() != 0;
 
 	if (!strcmp(argv[optind], "exec")) {
 		if (!pid)
 			pid = tree_id; /* old usage */
 		if (!pid)
 			goto opt_pid_missing;
-		return cr_exec(pid, argv + optind + 1);
+		return cr_exec(pid, argv + optind + 1) != 0;
 	}
 
 	if (!strcmp(argv[optind], "page-server"))
-		return cr_page_server(opts.restore_detach);
+		return cr_page_server(opts.restore_detach) > 0 ? 0 : 1;
 
 	if (!strcmp(argv[optind], "service"))
 		return cr_service(opts.restore_detach);
 
-	pr_msg("Unknown command \"%s\"\n", argv[optind]);
+	if (!strcmp(argv[optind], "dedup"))
+		return cr_dedup() != 0;
+
+	pr_msg("Error: unknown command: %s\n", argv[optind]);
 usage:
 	pr_msg("\n"
 "Usage:\n"
@@ -370,6 +379,7 @@ usage:
 "  criu exec -p PID <syscall-string>\n"
 "  criu page-server\n"
 "  criu service [<options>]\n"
+"  criu dedup\n"
 "\n"
 "Commands:\n"
 "  dump           checkpoint a process/tree identified by pid\n"
@@ -380,11 +390,12 @@ usage:
 "  exec           execute a system call by other task\n"
 "  page-server    launch page server\n"
 "  service        launch service\n"
+"  dedup          remove duplicates in memory dump\n"
 	);
 
-	if (argc < 2) {
+	if (usage_error) {
 		pr_msg("\nTry -h|--help for more info\n");
-		return -1;
+		return 1;
 	}
 
 	pr_msg("\n"
@@ -396,10 +407,9 @@ usage:
 "  -s|--leave-stopped    leave tasks in stopped state after checkpoint\n"
 "  -R|--leave-running    leave tasks in running state after checkpoint\n"
 "  -D|--images-dir DIR   directory for image files\n"
-"     --pidfile FILE     write a pid of a root task, service or page-server\n"
-"                        to this file\n"
-"  -W|--work-dir DIR     directory for logs/pidfiles/stats and for criu process itself\n"
-"                        if not specified, the --images-dir is used\n"
+"     --pidfile FILE     write root task, service or page-server pid to FILE\n"
+"  -W|--work-dir DIR     directory to cd and write logs/pidfiles/stats to\n"
+"                        (if not specified, value of --images-dir is used)\n"
 "\n"
 "* Special resources support:\n"
 "  -x|--" USK_EXT_PARAM "      allow external unix connections\n"
@@ -412,29 +422,32 @@ usage:
 "  --action-script FILE  add an external action script\n"
 "  -j|--" OPT_SHELL_JOB "        allow to dump and restore shell jobs\n"
 "  -l|--" OPT_FILE_LOCKS "       handle file locks, for safety, only used for container\n"
+"  -L|--libdir           path to a plugin directory (by default " CR_PLUGIN_DEFAULT ")\n"
 "\n"
 "* Logging:\n"
 "  -o|--log-file FILE    log file name\n"
 "     --log-pid          enable per-process logging to separate FILE.pid files\n"
-"  -v[NUM]               set logging level:\n"
-"                          -v0        - messages regardless of log level\n"
-"                          -v1, -v    - errors, when we are in trouble\n"
-"                          -v2, -vv   - warnings (default)\n"
-"                          -v3, -vvv  - informative, everything is fine\n"
-"                          -v4, -vvvv - debug only\n"
+"  -v[NUM]               set logging level (higher level means more output):\n"
+"                          -v1|-v    - only errors and messages\n"
+"                          -v2|-vv   - also warnings (default level)\n"
+"                          -v3|-vvv  - also information messages and timestamps\n"
+"                          -v4|-vvvv - lots of debug\n"
 "\n"
 "* Memory dumping options:\n"
 "  --track-mem           turn on memory changes tracker in kernel\n"
 "  --prev-images-dir DIR path to images from previous dump (relative to -D)\n"
 "  --page-server         send pages to page server (see options below as well)\n"
+"  --auto-dedup          when used on dump it will deduplicate \"old\" data in\n"
+"                        pages images of previous dump\n"
 "\n"
-"Page/Service server options\n"
+"Page/Service server options:\n"
 "  --address ADDR        address of server or service\n"
 "  --port PORT           port of page server\n"
 "  -d|--daemon           run in the background after creating socket\n"
 "\n"
 "Show options:\n"
 "  -f|--file FILE        show contents of a checkpoint file\n"
+"  -F|--fields FIELDS    show specified fields (comma separated)\n"
 "  -D|--images-dir DIR   directory where to get images from\n"
 "  -c|--contents         show contents of pages dumped in hexdump format\n"
 "  -p|--pid PID          show files relevant to PID (filter -D flood)\n"
@@ -445,9 +458,18 @@ usage:
 "     --ms               don't check not yet merged kernel features\n"
 	);
 
-	return -1;
+	return 0;
 
 opt_pid_missing:
-	pr_msg("No pid specified (-t option missing)\n");
-	return -1;
+	pr_msg("Error: pid not specified\n");
+	return 1;
+
+bad_arg:
+	if (idx < 0) /* short option */
+		pr_msg("Error: invalid argument for -%c: %s\n",
+				opt, optarg);
+	else /* long option */
+		pr_msg("Error: invalid argument for --%s: %s\n",
+				long_opts[idx].name, optarg);
+	return 1;
 }
