@@ -4,7 +4,7 @@
 #include <fcntl.h>
 #include <stdlib.h>
 
-#include "crtools.h"
+#include "fdset.h"
 #include "image.h"
 #include "files.h"
 #include "files-reg.h"
@@ -33,6 +33,7 @@ struct fifo_info {
 	struct file_desc	d;
 	FifoEntry		*fe;
 	bool			restore_data;
+	struct file_desc	*reg_d;
 };
 
 static LIST_HEAD(fifo_head);
@@ -109,12 +110,24 @@ static int open_fifo_fd(struct file_desc *d)
 {
 	struct fifo_info *info = container_of(d, struct fifo_info, d);
 
-	return open_path_by_id(info->fe->id, do_open_fifo, info);
+	return open_path(info->reg_d, do_open_fifo, info);
+}
+
+static void collect_fifo_fd(struct file_desc *d,
+		struct fdinfo_list_entry *fle, struct rst_info *ri)
+{
+	struct fifo_info *info;
+
+	info = container_of(d, struct fifo_info, d);
+	info->reg_d = collect_special_file(info->fe->id);
+	BUG_ON(info->reg_d == NULL);
+	collect_gen_fd(fle, ri);
 }
 
 static struct file_desc_ops fifo_desc_ops = {
 	.type		= FD_TYPES__FIFO,
 	.open		= open_fifo_fd,
+	.collect_fd	= collect_fifo_fd,
 };
 
 static int collect_one_fifo(void *o, ProtobufCMessage *base)
