@@ -33,6 +33,7 @@
 
 #define USK_EXTERN	(1 << 0)
 #define USK_SERVICE	(1 << 1)
+#define USK_CALLBACK	(1 << 2)
 
 #define VMA_AREA_NONE		(0 <<  0)
 #define VMA_AREA_REGULAR	(1 <<  0)	/* Dumpable area */
@@ -64,7 +65,7 @@
 extern bool fdinfo_per_id;
 extern bool ns_per_id;
 
-#define O_DUMP	(O_RDWR | O_CREAT | O_EXCL)
+#define O_DUMP	(O_RDWR | O_CREAT | O_TRUNC)
 #define O_SHOW	(O_RDONLY)
 #define O_RSTR	(O_RDONLY)
 
