@@ -84,17 +84,19 @@ struct proc_status_creds {
 	unsigned int uids[4];
 	unsigned int gids[4];
 
-	uint32_t cap_inh[PROC_CAP_SIZE];
-	uint32_t cap_prm[PROC_CAP_SIZE];
-	uint32_t cap_eff[PROC_CAP_SIZE];
-	uint32_t cap_bnd[PROC_CAP_SIZE];
+	u32 cap_inh[PROC_CAP_SIZE];
+	u32 cap_prm[PROC_CAP_SIZE];
+	u32 cap_eff[PROC_CAP_SIZE];
+	u32 cap_bnd[PROC_CAP_SIZE];
 };
 
 struct mount_info;
 struct fstype {
 	char *name;
+	int code;
 	int (*dump)(struct mount_info *pm);
 	int (*restore)(struct mount_info *pm);
+	int (*parse)(struct mount_info *pm);
 };
 
 struct mount_info {
@@ -110,6 +112,8 @@ struct mount_info {
 	char		*source;
 	char		*options;
 	bool		mounted;
+	bool		need_plugin;
+	int		is_file;
 	struct mount_info *next;
 
 	/* tree linkage */
@@ -125,6 +129,8 @@ struct mount_info {
 	struct mount_info *mnt_master;	/* slave is on master->mnt_slave_list */
 
 	struct list_head postpone;
+
+	void		*private;	/* associated filesystem data */
 };
 
 extern struct mount_info *mnt_entry_alloc();
@@ -136,6 +142,7 @@ extern struct mount_info *parse_mountinfo(pid_t pid);
 extern int parse_pid_stat(pid_t pid, struct proc_pid_stat *s);
 extern int parse_pid_stat_small(pid_t pid, struct proc_pid_stat_small *s);
 extern int parse_smaps(pid_t pid, struct vm_area_list *vma_area_list, bool use_map_files);
+extern int parse_self_maps_lite(struct vm_area_list *vms);
 extern int parse_pid_status(pid_t pid, struct proc_status_creds *);
 
 union fdinfo_entries {
@@ -148,6 +155,8 @@ union fdinfo_entries {
 
 extern int parse_fdinfo(int fd, int type,
 		int (*cb)(union fdinfo_entries *e, void *arg), void *arg);
+extern int parse_fdinfo_pid(int pid, int fd, int type,
+		int (*cb)(union fdinfo_entries *e, void *arg), void *arg);
 extern int parse_cpuinfo_features(int (*handler)(char *tok));
 extern int parse_file_locks(void);
 
