@@ -6,7 +6,7 @@
 
 enum {
 	/* PB_AUTOGEN_START */
-	PB_INVENTORY,
+	PB_INVENTORY,		/* 0 */
 	PB_STATS,
 	PB_FDINFO,
 	PB_CORE,
@@ -16,7 +16,7 @@ enum {
 	PB_POSIX_TIMER,
 	PB_CREDS,
 	PB_FS,
-	PB_UTSNS,
+	PB_UTSNS,		/* 10 */
 	PB_IPC_VAR,
 	PB_IPC_SHM,
 	PB_IPC_SEM,
@@ -25,7 +25,8 @@ enum {
 	PB_GHOST_FILE,
 	PB_TCP_STREAM,
 	PB_REG_FILE,
-	PB_NS_FILE,
+	PB_EXT_FILE,
+	PB_NS_FILE,		/* 20 */
 	PB_INET_SK,
 	PB_UNIX_SK,
 	PB_PACKET_SOCK,
@@ -35,7 +36,7 @@ enum {
 	PB_PIPE_DATA,
 	PB_EVENTFD_FILE,
 	PB_EVENTPOLL_FILE,
-	PB_EVENTPOLL_TFD,
+	PB_EVENTPOLL_TFD,	/* 30 */
 	PB_SIGNALFD,
 	PB_INOTIFY_FILE,
 	PB_INOTIFY_WD,
@@ -45,9 +46,10 @@ enum {
 	PB_TTY_INFO,
 	PB_FILE_LOCK,
 	PB_RLIMIT,
-	PB_PAGEMAP,
+	PB_PAGEMAP,		/* 40 */
 	PB_SIGINFO,
 	PB_TUNFILE,
+	PB_IRMAP_CACHE,
 
 	/* PB_AUTOGEN_STOP */
 
@@ -57,7 +59,7 @@ enum {
 	PB_NETDEV,
 	PB_REMAP_FPATH,
 	PB_SK_QUEUES,
-	PB_IPCNS_MSG,
+	PB_IPCNS_MSG,		/* 50 */
 	PB_IPCNS_MSG_ENT,
 
 	PB_MAX,
