@@ -9,11 +9,10 @@ struct cr_fdset;
 struct fd_parms;
 struct vma_area;
 
-int collect_packet_sockets(void);
-void show_packetsk(int fd);
+extern struct collect_image_info packet_sk_cinfo;
 
-int dump_socket_map(struct vma_area *vma);
-int get_socket_fd(int pid, VmaEntry *vma);
+extern int dump_socket_map(struct vma_area *vma);
+extern int get_socket_fd(int pid, VmaEntry *vma);
 
 extern int packet_receive_one(struct nlmsghdr *h, void *arg);
 
@@ -23,7 +22,9 @@ extern int packet_receive_one(struct nlmsghdr *h, void *arg);
 
 #ifndef PACKET_FANOUT
 #define PACKET_FANOUT	18
+#endif
 
+#ifndef TPACKET3_HDRLEN
 struct tpacket_req3 {
 	unsigned int tp_block_size;
 	unsigned int tp_block_nr;
