@@ -246,7 +246,7 @@ static int get_sockaddr_in(struct sockaddr_in *addr)
 
 int cr_page_server(bool daemon_mode)
 {
-	int sk, ask = -1;
+	int sk, ask = -1, ret;
 	struct sockaddr_in saddr, caddr;
 	socklen_t clen = sizeof(caddr);
 
@@ -273,11 +273,15 @@ int cr_page_server(bool daemon_mode)
 		goto out;
 	}
 
-	if (daemon_mode)
-		if (daemon(1, 0) == -1) {
+	if (daemon_mode) {
+		ret = cr_daemon(1, 0);
+		if (ret == -1) {
 			pr_perror("Can't run in the background");
-			return -errno;
+			goto out;
 		}
+		if (ret > 0) /* parent task, daemon started */
+			return ret;
+	}
 
 	if (opts.pidfile) {
 		if (write_pidfile(getpid()) == -1) {
@@ -286,21 +290,28 @@ int cr_page_server(bool daemon_mode)
 		}
 	}
 
-	ask = accept(sk, (struct sockaddr *)&caddr, &clen);
+	ret = ask = accept(sk, (struct sockaddr *)&caddr, &clen);
 	if (ask < 0)
 		pr_perror("Can't accept connection to server");
 
-out:
 	close(sk);
 
-	if (ask < 0)
-		return -1;
+	if (ask >= 0) {
+		pr_info("Accepted connection from %s:%u\n",
+				inet_ntoa(caddr.sin_addr),
+				(int)ntohs(caddr.sin_port));
 
-	pr_info("Accepted connection from %s:%u\n",
-			inet_ntoa(caddr.sin_addr),
-			(int)ntohs(caddr.sin_port));
+		ret = page_server_serve(ask);
+	}
 
-	return page_server_serve(ask);
+	if (daemon_mode)
+		exit(ret);
+
+	return ret;
+
+out:
+	close(sk);
+	return -1;
 }
 
 static int page_server_sk = -1;
@@ -462,8 +473,16 @@ static int write_pagemap_loc(struct page_xfer *xfer,
 static int write_pages_loc(struct page_xfer *xfer,
 		int p, unsigned long len)
 {
-	if (splice(p, NULL, xfer->fd_pg, NULL, len, SPLICE_F_MOVE) != len)
+	ssize_t ret;
+	ret = splice(p, NULL, xfer->fd_pg, NULL, len, SPLICE_F_MOVE);
+	if (ret == -1) {
+		pr_perror("Unable to spice data");
 		return -1;
+	}
+	if (ret != len) {
+		pr_err("Only %zu of %lu bytes have been spliced\n", ret, len);
+		return -1;
+	}
 
 	return 0;
 }
@@ -538,6 +557,11 @@ static int write_pagehole_loc(struct page_xfer *xfer, struct iovec *iov)
 
 static void close_page_xfer(struct page_xfer *xfer)
 {
+	if (xfer->parent != NULL) {
+		xfer->parent->close(xfer->parent);
+		xfree(xfer->parent);
+		xfer->parent = NULL;
+	}
 	close(xfer->fd_pg);
 	close(xfer->fd);
 }
@@ -617,7 +641,8 @@ static int open_page_local_xfer(struct page_xfer *xfer, int fd_type, long id)
 	 * 2) when writing a hole, the respective place would be checked
 	 *    to exist in parent (either pagemap or hole)
 	 */
-	{
+	xfer->parent = NULL;
+	if (fd_type == CR_FD_PAGEMAP) {
 		int ret;
 		int pfd;
 
