@@ -160,7 +160,6 @@ static int show_bytes(pb_pr_field_t *field)
 static size_t pb_show_prepare_field_context(const ProtobufCFieldDescriptor *fd,
 					  pb_pr_ctl_t *ctl)
 {
-	pb_pr_field_t *field = &ctl->cur;
 	size_t fsize = 0;
 
 	switch (fd->type) {
@@ -184,7 +183,6 @@ static size_t pb_show_prepare_field_context(const ProtobufCFieldDescriptor *fd,
 		break;
 	case PROTOBUF_C_TYPE_MESSAGE:
 		ctl->arg = (void *)fd->descriptor;
-		field->data = (void *)(*(long *)field->data);
 	case PROTOBUF_C_TYPE_STRING:
 		fsize = sizeof (void *);
 		break;
@@ -348,7 +346,7 @@ static pb_pr_show_t get_show_function(const ProtobufCFieldDescriptor *fd, pb_pr_
 	return get_pb_show_function(fd->type, fd->label);
 }
 
-static void pb_show_repeated(pb_pr_ctl_t *ctl, int nr_fields, pb_pr_show_t show,
+static void pb_show_repeated(const ProtobufCFieldDescriptor *fd, pb_pr_ctl_t *ctl, int nr_fields, pb_pr_show_t show,
 			  size_t fsize)
 {
 	pb_pr_field_t *field = &ctl->cur;
@@ -360,6 +358,23 @@ static void pb_show_repeated(pb_pr_ctl_t *ctl, int nr_fields, pb_pr_show_t show,
 		return;
 	}
 
+	if (fd->type == PROTOBUF_C_TYPE_MESSAGE) {
+		void *p = field->data;
+
+		field->count = nr_fields;
+		field->data = (void *)(*(long *)p);
+		done = show(field);
+		if (done)
+			return;
+
+		for (p += fsize, counter = 0; counter < nr_fields - 1; counter++, p += fsize) {
+			pr_msg(":");
+			field->data = (void *)(*(long *)p);
+			show(field);
+		}
+		return;
+	}
+
 	field->count = nr_fields;
 	done = show(field);
 	if (done)
@@ -383,7 +398,7 @@ static void pb_show_field(const ProtobufCFieldDescriptor *fd,
 
 	show = get_show_function(fd, ctl);
 
-	pb_show_repeated(ctl, nr_fields, show, pb_show_prepare_field_context(fd, ctl));
+	pb_show_repeated(fd, ctl, nr_fields, show, pb_show_prepare_field_context(fd, ctl));
 
 	if (ctl->single_entry)
 		pr_msg("\n");
@@ -599,6 +614,7 @@ int pb_write_one(int fd, void *obj, int type)
 	void *buf = (void *)&local;
 	u32 size, packed;
 	int ret = -1;
+	struct iovec iov[2];
 
 	if (!cr_pb_descs[type].pb_desc) {
 		pr_err("Wrong object requested %d\n", type);
@@ -618,17 +634,14 @@ int pb_write_one(int fd, void *obj, int type)
 		goto err;
 	}
 
-	ret = write(fd, &size, sizeof(size));
-	if (ret != sizeof(size)) {
-		ret = -1;
-		pr_perror("Can't write %d bytes", (int)sizeof(size));
-		goto err;
-	}
+	iov[0].iov_base = &size;
+	iov[0].iov_len = sizeof(size);
+	iov[1].iov_base = buf;
+	iov[1].iov_len = size;
 
-	ret = write(fd, buf, size);
-	if (ret != size) {
-		ret = -1;
-		pr_perror("Can't write %d bytes", size);
+	ret = writev(fd, iov, 2);
+	if (ret != size + sizeof(size)) {
+		pr_perror("Can't write %d bytes", (int)(size + sizeof(size)));
 		goto err;
 	}
 
