@@ -38,7 +38,7 @@ static struct shmem_info *find_shmem_by_id(unsigned long id)
 	return find_shmem(si, nr_shmems, id);
 }
 
-static int collect_shmem(int pid, VmaEntry *vi)
+int collect_shmem(int pid, VmaEntry *vi)
 {
 	unsigned long size = vi->pgoff + vi->end - vi->start;
 	struct shmem_info *si;
@@ -85,43 +85,6 @@ static int collect_shmem(int pid, VmaEntry *vi)
 	return 0;
 }
 
-int prepare_shmem_pid(int pid)
-{
-	int fd, ret = -1;
-	VmaEntry *vi;
-
-	fd = open_image(CR_FD_VMAS, O_RSTR, pid);
-	if (fd < 0) {
-		if (errno == ENOENT)
-			return 0;
-		else
-			return -1;
-	}
-
-	while (1) {
-		ret = pb_read_one_eof(fd, &vi, PB_VMA);
-		if (ret <= 0)
-			break;
-
-		pr_info("vma 0x%"PRIx64" 0x%"PRIx64"\n", vi->start, vi->end);
-
-		if (!vma_entry_is(vi, VMA_ANON_SHARED) ||
-		    vma_entry_is(vi, VMA_AREA_SYSVIPC)) {
-			vma_entry__free_unpacked(vi, NULL);
-			continue;
-		}
-
-		ret = collect_shmem(pid, vi);
-		vma_entry__free_unpacked(vi, NULL);
-
-		if (ret)
-			break;
-	}
-
-	close(fd);
-	return ret;
-}
-
 static int shmem_wait_and_open(int pid, struct shmem_info *si)
 {
 	char path[128];
@@ -309,11 +272,25 @@ int add_shmem_area(pid_t pid, VmaEntry *vma)
 	return 0;
 }
 
+static int dump_pages(struct page_pipe *pp, struct page_xfer *xfer, void *addr)
+{
+	struct page_pipe_buf *ppb;
+
+	list_for_each_entry(ppb, &pp->bufs, l)
+		if (vmsplice(ppb->p[1], ppb->iov, ppb->nr_segs,
+					SPLICE_F_GIFT | SPLICE_F_NONBLOCK) !=
+				ppb->pages_in * PAGE_SIZE) {
+			pr_perror("Can't get shmem into page-pipe");
+			return -1;
+		}
+
+	return page_xfer_dump_pages(xfer, pp, (unsigned long)addr);
+}
+
 static int dump_one_shmem(struct shmem_info_dump *si)
 {
 	struct iovec *iovs;
 	struct page_pipe *pp;
-	struct page_pipe_buf *ppb;
 	struct page_xfer xfer;
 	int err, ret = -1, fd;
 	unsigned char *map = NULL;
@@ -354,32 +331,32 @@ static int dump_one_shmem(struct shmem_info_dump *si)
 	if (!iovs)
 		goto err_unmap;
 
-	pp = create_page_pipe((nrpages + 1) / 2, iovs);
+	pp = create_page_pipe((nrpages + 1) / 2, iovs, true);
 	if (!pp)
 		goto err_iovs;
 
-	for (pfn = 0; pfn < nrpages; pfn++) {
-		if (!(map[pfn] & PAGE_RSS))
-			continue;
-
-		if (page_pipe_add_page(pp, (unsigned long)addr + pfn * PAGE_SIZE))
-			goto err_pp;
-	}
-
-	list_for_each_entry(ppb, &pp->bufs, l)
-		if (vmsplice(ppb->p[1], ppb->iov, ppb->nr_segs,
-					SPLICE_F_GIFT | SPLICE_F_NONBLOCK) !=
-				ppb->pages_in * PAGE_SIZE) {
-			pr_perror("Can't get shmem into page-pipe");
-			goto err_pp;
-		}
-
 	err = open_page_xfer(&xfer, CR_FD_SHMEM_PAGEMAP, si->shmid);
 	if (err)
 		goto err_pp;
 
-	ret = page_xfer_dump_pages(&xfer, pp, (unsigned long)addr);
+	for (pfn = 0; pfn < nrpages; pfn++) {
+		if (!(map[pfn] & PAGE_RSS))
+			continue;
+again:
+		ret = page_pipe_add_page(pp, (unsigned long)addr + pfn * PAGE_SIZE);
+		if (ret == -EAGAIN) {
+			ret = dump_pages(pp, &xfer, addr);
+			if (ret)
+				goto err_xfer;
+			page_pipe_reinit(pp);
+			goto again;
+		} else if (ret)
+			goto err_xfer;
+	}
 
+	ret = dump_pages(pp, &xfer, addr);
+
+err_xfer:
 	xfer.close(&xfer);
 err_pp:
 	destroy_page_pipe(pp);
