@@ -93,6 +93,8 @@ static int prepare_mntns()
 	mkdir("/dev/net", 0777);
 	mknod("/dev/net/tun", 0777 | S_IFCHR, makedev(10, 200));
 	chmod("/dev/net/tun", 0777);
+	mknod("/dev/rtc", 0777 | S_IFCHR, makedev(254, 0));
+	chmod("/dev/rtc", 0777);
 	return 0;
 }
 
@@ -143,7 +145,7 @@ static void ns_sig_hand(int signo)
 	return;
 write_out:
 	/* fprintf can't be used in a sighandler due to glibc locks */
-	write(STDERR_FILENO, buf, MAX(len, sizeof(buf)));
+	write(STDERR_FILENO, buf, MIN(len, sizeof(buf)));
 }
 
 #define STATUS_FD 255
@@ -197,10 +199,6 @@ int ns_init(int argc, char **argv)
 		fprintf(stderr, "Can't set SIGTERM handler: %m\n");
 		exit(1);
 	}
-	if (sigaction(SIGCHLD, &sa, NULL)) {
-		fprintf(stderr, "Can't set SIGCHLD handler: %m\n");
-		exit(1);
-	}
 
 	/* Start test */
 	pid = fork();
@@ -212,9 +210,12 @@ int ns_init(int argc, char **argv)
 		fprintf(stderr, "exec(%s) failed: %m\n", argv[0]);
 		return ret;
 	}
-	ret = 1;
-	waitpid(pid, &ret, 0);
 
+	ret = -1;
+	if (waitpid(pid, &ret, 0) < 0)
+		fprintf(stderr, "waitpid() failed: %m\n");
+	else if (ret)
+		fprintf(stderr, "The test returned non-zero code %d\n", ret);
 
 	pid = fork();
 	if (pid == 0) {
@@ -224,6 +225,25 @@ int ns_init(int argc, char **argv)
 	} else if (pid > 0)
 		waitpid(pid, NULL, 0);
 
+	if (sigaction(SIGCHLD, &sa, NULL)) {
+		fprintf(stderr, "Can't set SIGCHLD handler: %m\n");
+		exit(1);
+	}
+
+	while (1) {
+		int status;
+
+		pid = waitpid(-1, &status, WNOHANG);
+		if (pid == 0)
+			break;
+		if (pid < 0) {
+			fprintf(stderr, "waitpid() failed: %m\n");
+			exit (1);
+		}
+		if (status)
+			fprintf(stderr, "%d return %d\n", pid, status);
+	}
+
 	/* Daemonize */
 	write(status_pipe, &ret, sizeof(ret));
 	close(status_pipe);
@@ -291,15 +311,21 @@ void ns_create(int argc, char **argv)
 
 	status = 1;
 	ret = read(args.status_pipe[0], &status, sizeof(status));
-	if (ret != sizeof(status) || status)
+	if (ret != sizeof(status) || status) {
+		fprintf(stderr, "The test failed (%d, %d)\n", ret, status);
 		exit(1);
+	}
 	ret = read(args.status_pipe[0], &status, sizeof(status));
-	if (ret != 0)
+	if (ret != 0) {
+		fprintf(stderr, "Unexpected message from test\n");
 		exit(1);
+	}
 
 	pidfile = getenv("ZDTM_PIDFILE");
-	if (pidfile == NULL)
+	if (pidfile == NULL) {
+		fprintf(stderr, "ZDTM_PIDFILE isn't defined");
 		exit(1);
+	}
 	fd = open(pidfile, O_CREAT | O_EXCL | O_WRONLY, 0666);
 	if (fd == -1) {
 		fprintf(stderr, "Can't create the file %s: %m\n", pidfile);
